#
# Ged2Wiki.pl - Perl program to convert a GEDCOM file to plain text Wiki pages
#
# The input format is based on "The GEDCOM Standard" Release 5.5, as published
# by the Family History Department of The Church of Jesus Christ of Latter-day
# Saints, 2 January 1996.
#
# The output format is a series of plain-text files based on DokuWiki
#
# For questions, comments, bug reports, etc., please visit this site:
#    http://s560.com/ged2wiki
#

# Constants used at various points - modify to suit your taste
my $publicbirth = 1900;    # People born AFTER this year are listed as private (unless dead)
my $publicmarr = 1920;     # People married AFTER this year are listed as private (unless dead)
my $fileprefix = "";       # Prefix for each wiki page filename
my $filesuffix = ".txt";   # Suffix for each wiki page filename
my $filenamelc = 1;        # Should we convert the filename to lower case?
my $contdotnewline = 1;    # Should we convert leading dot[space] to newline?
my $unknownsurname = "---";   # What should we show when the surname is unknown?

# Format styles (uses FMT_* constant below - suit to taste)
my $surnameformat = 1;     # Format for all surnames
my $titleformat = 0;       # Format for names in the Title line
my $linkformat = 0;        # Format for names that are used as links
my $normalformat = 2;      # Format for names that are NOT used as links

# Format constants (do not modify)
use constant FMT_UPPERCASE => 1;    # Make string UPPERCASE
use constant FMT_BOLD => 2;         # Make string **Bold**
use constant FMT_ITALICS => 4;      # Make string //Italics//
use constant FMT_UNDERLINE => 8;    # Make string __Underlined__

# Set the Debug output device here
#open DEBUG, ">&STDOUT";
open DEBUG, ">NUL";

# The globals for the GEDCOM input lines
my $currline;
my $nextline;
my $currlevel;
my $nextlevel;

# Global counters
my $famcount;
my $indicount;
my $wikicount;

# The globals for the decoded data about individuals
# (Index = Individual Xref tag)
my %name;         # Full Name
my %sex;          # 'M' or 'F'
my %birth;        # Birth Date and/or Place
my %death;        # Death Date and/or Place
my %burial;       # Burial Date and/or Place
my %familydown;   # Xref tag for Individual's Family (FAMS)
my %familyup;     # Xref tag for Parent's Family (FAMC)
my %inote;        # Note
my %private;      # Is data private? (If so, do not show details on output)
my %alone;        # Does this person need his/her own page?

# most of the data above can have an associated note
my %nnote;        # Note for Name
my %bnote;        # Note for Birth
my %dnote;        # Note for Death
my %burnote;      # Note for Burial
my %fdnote;       # Note for Family (Up)
my %funote;       # Note for Family (Down)

# The globals for the decoded data about families
# (Index = Family Xref tag)
my %familyflag;   # This entry will exist for ALL families
my %husband;      # Individual tag for husband
my %wife;         # Individual tag for wife
my %marriage;     # Marriage Date and/or Place
my %fnote;        # Note
my %children;     # List of individual tags for children
my %mnote;        # Note for Marriage event

# The global for records other than Individuals and Families
my %note;         # Note

# Prime the parser by reading the first line
fetch_line();

# "Read" the first line
read_line();

# The first line must be "0 HEAD"
if ($currline !~ /^HEAD/) {
   print "Input does not appear to be a GEDCOM file\n";
   exit;
}

# First line is good.  Parse the rest of the header
parse_header();

# If it's a Submission record, parse it here
if ($currline =~ /^\@(.*?)\@\s+SUBN/) {
   parse_subn($1);
}

# Loop here, reading the "Record" entries until we find the Trailer
while (1) {

   # First, check for the Trailer record
   if ($currline =~ /^TRLR/) {
      last; # We're done with the input - break out of this loop
   }

   # Each Record must be in this format
   if ($currline !~ /^\@(.*?)\@\s+(\w+)\s*/) {
      print "Unexpected Record at line $.:\n>> $currline\n";
      exit;
   }
   my $xref = $1;
   my $tag = $2;
   my $data = "$'";

   # If it's a "FAM" record, parse it here
   if ($tag eq "FAM") {
      parse_family($xref);
      next;
   }

   # If it's a "INDI" record, parse it here
   if ($tag eq "INDI") {
      parse_individual($xref);
      next;
   }

   # If it's a "NOTE" record, parse it here
   if ($tag eq "NOTE") {
      parse_note($xref, $data);
      next;
   }

   # If it's an unknown record, it's an error
   if ($tag !~ /^(OBJE|REPO|SOUR|SUBM)$/) {
      print "Unexpected Tag at line $.:\n>> $currline\n";
      exit;
   }

   # Ignore this type of record
   print DEBUG "Ignoring $currlevel " . substr($currline, 0, 60) . "\n";
   parse_ignore();
}

print "GEDCOM file parsed successfully:\n";
print "   $indicount Individuals\n";
print "   $famcount Families\n";

# Generate all the output here ...
foreach my $fxref (keys %familyflag) {
   # Output data for family $fxref
   output_family($fxref);
   $wikicount++;
}
foreach my $ixref (keys %name) {
   if ($alone{$ixref} == 1) {
      output_individual($ixref);
      $wikicount++;
   }
}
print "   $wikicount Wiki Pages created\n";
exit;

# Read one line from the input file (and strip unwanted whitespace)
sub fetch_line {
   $nextline = <>;
   $nextline =~ s/^\s+//;        # Strip leading whitespace
   $nextline =~ s/[\s\r\n]*$//;  # Strip trailing whitespace (and newline)
   if ($nextline !~ /^(\d\d?)\s+/) {
      print "Invalid GEDCOM data at line $.:\n>> $nextline\n";
      exit;
   }
   $nextlevel = $1;
   $nextline = "$'";
}

# One input line -> $currline (concatenating as needed)
sub read_line {

   $currline = $nextline;
   $currlevel = $nextlevel;

   # Special case: Do not read beyond the Trailer
   if ($currline =~ /^TRLR/) {
      $nextline = "";
      $nextlevel = "";
      return;
   }

   fetch_line();

   # Special case: A tag without any parameters - Add a single space
   if ($currline =~ /^(\@\w+\@)?\s+\w+$/) {
      $currline .= " ";
   }

   while ($nextlevel == $currlevel + 1 && $nextline =~ /^CON([CT])\s?/) {
      my $rlcont = "$'";
      if ($1 eq "C") {
         $currline .= $rlcont;
      } else {
         if ($contdotnewline != 0) {
            $rlcont =~ s/^(\.\s)+/\n/; # Any number of ". " -> Newline
            $rlcont =~ s/^\.$/\n/;     # If entire line is "." -> Newline
         }
         $currline .= "\n$rlcont";
      }
      fetch_line();
   }
}

# Parse the header
sub parse_header {
   # For now, ignore this type of record
   print DEBUG "Ignoring $currlevel " . substr($currline, 0, 60) . "\n";
   parse_ignore();

   # Format of Header (brackets denote optional entries)
   #   0 HEAD
   #   1 SOUR
   # [ 1 DEST ]
   # [ 1 DATE ]
   #   1 SUBM
   # [ 1 SUBN ]
   # [ 1 FILE ]
   # [ 1 COPR ]
   #   1 GEDC
   #   1 CHAR
   # [ 1 LANG ]
   # [ 1 PLAC ]
   # [ 1 NOTE ]

}

# Parse the submission record
sub parse_subn {
   # For now, ignore this type of record
   print DEBUG "Ignoring $currlevel " . substr($currline, 0, 60) . "\n";
   parse_ignore();

   # Format of Submission record (all sub-fields are optional)
   #   0 @XREF@ SUBN
   # [ 1 SUBM ]
   # [ 1 FAMF ]
   # [ 1 TEMP ]
   # [ 1 ANCE ]
   # [ 1 DESC ]
   # [ 1 ORDI ]
   # [ 1 RIN ]

}

# Parse a Family record
# Parm1 = Xref id for Family
sub parse_family {
   my $ftag = shift;

   # All families must have this marker entry
   $familyflag{$ftag} = 1;

   # Increment our counter
   $famcount++;

   # Print a debug line
   print DEBUG "Family $ftag\n";

   # Format of Family record (brackets denote optional entries)
   #   0 @XREF@ FAM
   # { 1 ANUL|CENS|DIV|DIVF|ENGA|MARB|MARC|MARL|MARS|EVEN }
   # { 1 MARR }   ->    %marriage
   # [ 1 HUSB ]   ->    %husband
   # [ 1 WIFE ]   ->    %wife
   # { 1 CHIL }   ->    %children
   # [ 1 NCHI ]
   # { 1 SUBM }
   # [ 1 SLGS ]
   # { 1 SOUR }
   # { 1 OBJE }
   # { 1 NOTE }   ->    %mnote
   # { 1 REFN }
   # [ 1 RIN ]
   # [ 1 CHAN ]

   # Read the first sub-line
   read_line();

   while ($currlevel > 0) {

      my $work;

      # Record this family's husband
      # (If there are multiple entries of this type, use only the first)
      if ($currline =~ /^HUSB\s+\@(.*?)\@/) {
         if ($husband{$ftag} eq "") {
            $husband{$ftag} = $1;
         }
      }

      # Record this family's wife
      # (If there are multiple entries of this type, use only the first)
      if ($currline =~ /^WIFE\s+\@(.*?)\@/) {
         if ($wife{$ftag} eq "") {
            $wife{$ftag} = $1;
         }
      }

      # Record this family's children
      # (The result is a comma-separated list of Individual tags)
      if ($currline =~ /^CHIL\s+\@(.*?)\@/) {
         if ($children{$ftag} eq "") {
            $children{$ftag} = $1;
         } else {
            $children{$ftag} .= ",$1";
         }
      }

      # Record this family's marriage event
      if ($currline =~ /^MARR/) {
         if ($marriage{$ftag} eq "") {
            $marriage{$ftag} = read_event();
            if ($currlevel == 0) {
               last;
            }
         }
         if ($currlevel > 1) {
            $work = read_note();
            if ($work ne "") {
               $mnote{$ftag} .= $work;
            }
         }
         next;
      }

      # Record any notes associated with this family
      if ($currline =~ /^NOTE\s+/) {
         $fnote{$ftag} .= "$'";
      }

      # If we get here, we need to ignore the remaining lines of this sub-record
##      print DEBUG "Ignoring $currlevel " . substr($currline, 0, 60) . "\n";
      parse_ignore();

   }

}

# Parse an Individual record
# Parm1 = Xref id for Individual
sub parse_individual {
   my $itag = shift;

   # Increment our counter
   $indicount++;

   # Print a debug line
   print DEBUG "Individual $itag\n";

   # Format of Individual record (brackets denote optional entries)
   #   0 @XREF@ INDI
   # { 1 NAME }   ->    %name;
   # [ 1 SEX ]    ->    %sex;
   # { 1 BIRT }   ->    %birth;
   # { 1 DEAT }   ->    %death;
   # { 1 BURI }   ->    %burial;
   # { 1 CHR|CREM|ADOP|BAPM|BARM|BASM|BLES|CHRA|CONF|FCOM|
   #     ORDN|NATU|EMIG|IMMI|CENS|PROB|WILL|GRAD|RETI|EVEN }
   # { 1 CAST|DSCR|EDUC|IDNO|NATI|NCHI|NMR|OCCU|PROP|RELI|RESI|SSN|TITL }
   # { 1 BAPL|CONL|ENDL|SLGC }
   # { 1 FAMC }   ->    &familyup;
   # { 1 FAMS }   ->    %familydown;
   # { 1 SUBM }
   # { 1 ASSO }
   # { 1 ALIA }
   # { 1 ANCI }
   # { 1 DESI }
   # { 1 SOUR }
   # { 1 OBJE }
   # { 1 NOTE }   ->    %inote;
   # [ 1 RFN ]
   # [ 1 AFN ]
   # { 1 REFN }
   # [ 1 RIN ]
   # [ 1 CHAN ]

   # Read the first sub-line
   read_line();

   # Start off with the assumption that this person's data will be private
   $private{$itag} = 1;

   # Also assume that this person has his/her own family
   $alone{$ctag} = 0;

   while ($currlevel > 0) {

      my $work;

      # Record this person's name
      # (If there are multiple entries of this type, use only the first)
      if ($currline =~ /^NAME\s+/) {
         if ($name{$itag} eq "") {
            $name{$itag} = "$'"; # Save the full name
         }
         $work = read_note();
         if ($work ne "") {
            $nnote{$itag} = $work;
         }
         next;
      }

      # Record this person's gender
      if ($currline =~ /^SEX\s+/) {
         if ($sex{$itag} eq "") {
            $sex{$itag} = uc substr("$'", 0, 1);
         }
         # Get the next line and loop...
         read_line();
         next;
      }

      # Record this person's birth date/place
      if ($currline =~ /^BIRT/) {
         if ($birth{$itag} eq "") {
            $birth{$itag} = read_event();
            if ($currlevel == 0) {
               last;
            }
         }
         if ($currlevel > 1) {
            $work = read_note();
            if ($work ne "") {
               $bnote{$itag} .= $work;
            }
         }
         # If they were born before a certain date, their data can be public
         if ($birth{$itag} =~ /(\d{4})/) {
            if ($1 <= $publicbirth) {
               $private{$itag} = 0;
            }
         }
         next;
      }

      # Record this person's death date/place
      if ($currline =~ /^DEAT/) {
         if ($death{$itag} eq "") {
            $death{$itag} = read_event();
            if ($currlevel == 0) {
               last;
            }
         }
         if ($currlevel > 1) {
            $work = read_note();
            if ($work ne "") {
               $dnote{$itag} .= $work;
            }
         }
         # If they're dead, their data can be public
         $private{$itag} = 0;
         next;
      }

      # Record this person's burial date/place
      if ($currline =~ /^BURI/) {
         if ($burial{$itag} eq "") {
            $burial{$itag} = read_event();
            if ($currlevel == 0) {
               last;
            }
         }
         if ($currlevel > 1) {
            $work = read_note();
            if ($work ne "") {
               $burnote{$itag} .= $work;
            }
         }
         # If they're buried, their data can be public
         $private{$itag} = 0;
         next;
      }

      # Record the family in which this person is a child
      # (There may be multiple entries of this type)
      if ($currline =~ /^FAMC\s+\@(.*?)\@/) {
         if ($familyup{$itag} eq "") {
            $familyup{$itag} = $1;
         } else {
            $familyup{$itag} .= ",$1";
         }
         $work = read_note();
         if ($work ne "") {
            $funote{$itag} .= $work;
         }
         next;
      }

      # Record the family in which this person is a spouse
      # (There may be multiple entries of this type)
      if ($currline =~ /^FAMS\s+\@(.*?)\@/) {
         if ($familydown{$itag} eq "") {
            $familydown{$itag} = $1;
         } else {
            $familydown{$itag} .= ",$1";
         }
         $work = read_note();
         if ($work ne "") {
            $fdnote{$itag} = $work;
         }
         next;
      }

      # Record any notes associated with this person
      if ($currline =~ /^NOTE\s+/) {
         $inote{$itag} .= "$'";
      }

      # If we get here, we need to ignore the remaining lines of this sub-record
##      print DEBUG "Ignoring $currlevel " . substr($currline, 0, 60) . "\n";
      parse_ignore();

   }

}

# Parse a Note record
# Parm1 = Xref id for Note
# Parm2 = Text of Note
sub parse_note {
   # Save the note's data
   $note{$_[0]} = $_[1];

   # Print a debug line
   print DEBUG "Note $_[0]\n";

   # Format of Note record
   #   0 @XREF@ NOTE
   # { 1 SOUR }
   # { 1 REFN }
   # [ 1 RIN ]
   # [ 1 CHAN ]

   # And ignore all sub-records
   parse_ignore();
}

# Read lines until we find a level the same or lower than the current line
# Examples:
# If the current line is "0 HEAD", it will read until it finds another "0" line.
# if the current line is "1 NAME", it will read until it finds a "1" or a "0" line.
# In either case, the "current" line will be the line AFTER the ignored block
sub parse_ignore {
   my $thislevel = $currlevel;

   while (1) {
      read_line();
      if ($currlevel <= $thislevel) {
         last;
      }
   }
}

# Read and parse an "Event Detail" entry
sub read_event {
   my $thislevel = $currlevel;
   my $edate;
   my $eplace;
   while (1) {
      read_line();
      if ($currlevel <= $thislevel) {
         last;
      }
      if ($currline =~ /^DATE\s+/) {
         $edate = "$'";
      }
      if ($currline =~ /^PLAC\s+/) {
         $eplace = "$'";
      }
   }
   if ($eplace ne "" && $edate ne "") {
      "$edate, $eplace";
   } else {
      if ($eplace ne "") {
         $eplace;
      } else {
         $edate;
      }
   }
}

# Read the sub-records, looking for a NOTE entry
sub read_note {
   my $thislevel = $currlevel;
   my $thisnote;
   while (1) {
      read_line();
      if ($currlevel <= $thislevel) {
         last;
      }
      if ($currline =~ /^NOTE\s+/) {
         $thisnote .= "$'";
      }
   }
   $thisnote;
}

# Find the first family for a given individual
# Parm1 = Xref id for Family
sub first_family {
   my $ftag = shift;

   # If the tag is a list, throw away all but the first
   if ($ftag =~ /\,/) {
      $ftag = $`;
   }

   # Return the tag for the first family
   $ftag;
}

# Find the first father of a given individual
# Parm1 = Xref id for Individual
sub first_father {
   # Get the family (or list of families) for this individual
   my $ftag = $familyup{$_[0]};

   # If there is no family, leave the result blank
   if ($ftag ne "") {
      # Otherwise, return the husband in the first family in the list
      $ftag = $husband{first_family($ftag)};
   }
   $ftag;
}

# Output a single family
# Parm1 = Xref id for Family to output
sub output_family {
   my $ftag = shift;
   my $fname = "$fileprefix$ftag$filesuffix";

   if ($filenamelc != 0) {
      $fname = lc $fname;
   }

   open FILE, ">$fname" or die $!;

   # Save the husband and wife's tag (since we're going to use them multiple times)
   my $htag = $husband{$ftag};
   my $wtag = $wife{$ftag};

   # Special Case: If the marriage date is older than a certain year, then make sure
   # neither spouses is marked as "private".
   if ($marriage{$ftag} =~ /(\d{4})/) {
      if ($1 <= $publicmarr) {
         $private{$htag} = 0;
         $private{$wtag} = 0;
      }
   }

   my $name1;
   my $name2;

   # Create the Page Title
   if ($htag ne "" && $wtag ne "") {
      $name1 = format_name($htag, $titleformat);
      $name2 = format_name($wtag, $titleformat);
      print FILE "==== $name1 & $name2 ====\n\n";
   } else {
      if ($htag eq "") {
         if ($wtag eq "") {
            print FILE "==== Unknown Parents ====\n\n";
         } else {
            $name2 = format_name($wtag, $titleformat);
            print FILE "==== Children of $name2 ====\n\n";
         }
      } else {
         $name1 = format_name($htag, $titleformat);
         print FILE "==== Children of $name1 ====\n\n";
      }
   }

   # Create the Husband entry (if present)
   if ($htag ne "") {
      $name1 = format_name($htag, $normalformat);
      if ($children{$ftag} ne "") {
         print FILE "Father: ";
      } else {
         print FILE "Husband: ";
      }
      print FILE "$name1";
      if ($familyup{$htag} ne "") {
         $name2 = format_name(first_father($htag), $linkformat);
         print FILE ",\n(son of ";
         print FILE format_link($name2, first_family($familyup{$htag}));
         print FILE ")";
      }
      print FILE format_data($htag);
      print FILE "\n\n";
      output_note($inote{$htag}, 0);
   }

   # Create the Wife entry (if present)
   if ($wtag ne "") {
      $name1 = format_name($wtag, $normalformat);
      if ($children{$ftag} ne "") {
         print FILE "Mother: ";
      } else {
         print FILE "Wife: ";
      }
      print FILE $name1;
      if ($familyup{$wtag} ne "") {
         $name2 = format_name(first_father($wtag), $linkformat);
         print FILE ",\n(daughter of ";
         print FILE format_link($name2, first_family($familyup{$wtag}));
         print FILE ")";
      }
      print FILE format_data($wtag);
      print FILE "\n\n";
      output_note($inote{$wtag}, 0);
   }

   # Show the marriage event information
   if ($marriage{$ftag} ne "") {
      print FILE "Married $marriage{$ftag}\n";
      output_note($mnote{$ftag}, 1);
   }

   # If there's a note for this family, show it here
   output_note($fnote{$ftag}, 0);

   # Create the entries for the children (if any exist)
   if ($children{$ftag} ne "") {
      my @children = split(/\,/, $children{$ftag});
      if ($#children > 0) {
         print FILE "Children:\n";
      } else {
         print FILE "Child:\n";
      }
      foreach my $ctag (@children) {
         # Format the name of the child
         $name1 = format_name($ctag, $linkformat);
         print FILE "  - ";
         if ($familydown{$ctag} eq "") {
            if ($private{$ctag} == 0) {
               print FILE format_link($name1, $ctag);
               $alone{$ctag} = 1;
            } else {
               print FILE $name1;
            }
         } else {
            print FILE format_link($name1, first_family($familydown{$ctag}));
         }
         # Show the gender even if private
         if ($sex{ctag} ne "") {
            $outstr = " ($sex{ctag})";
         }
         print FILE format_data($ctag);
         print FILE "\n";
      }
   }

   close FILE;
}

# Output a single individual (only for people without a family)
# Parm1 = Xref id for Individual to output
sub output_individual {
   my $itag = shift;
   my $fname = "$fileprefix$itag$filesuffix";

   if ($filenamelc != 0) {
      $fname = lc $fname;
   }

   open FILE, ">$fname" or die $!;

   my $name1;

   # Create the Page Title
   $name1 = format_name($itag, $titleformat);
   print FILE "==== $name1 ====\n\n";

   # Create the Individual entry
   $name1 = format_name($itag, $normalformat);
   print FILE $name1;
   if ($familyup{$itag} ne "") {
      $name2 = format_name(first_father($itag), $linkformat);
      print FILE ",\n(";
      if ($sex{$itag} eq "M") {
         print FILE "son";
      } else {
         if ($sex{$itag} eq "F") {
            print FILE "daughter";
         } else {
            print FILE "child";
         }
      }
      print FILE " of ";
      print FILE format_link($name2, first_family($familyup{$itag}));
      print FILE ")";
   }
   print FILE format_data($itag);
   print FILE "\n\n";

   # If there's a note for this person, show it here
   output_note($inote{$itag}, 0);

   close FILE;
}

# Format a name according to the specified rules
# Parm1 = Xref id for Individual
# Parm2 = Format option for output
sub format_name {
   my $itag = shift;          # The xref id
   my $nfmt = shift;          # The format option for name
   my $nstr = $name{$itag};   # Look up the name from the xref id
   my $n1, $lname, $n2;

   # The surname format options are the bits in $surnameformat that are NOT also in $nfmt
   my $lfmt = $surnameformat & ~$nfmt;

   # Extract the surname (which is surrounded by slashes)
   if ($nstr =~ /\/\s*(.*?)\s*\//) {
      $n1 = "$`";       # The given name(s)
      $lname = $1;      # No leading/trailing spaces on surname
      $n2 = "$'";       # Whatever's after the surname

      $n1 =~ s/\s+$//;  # Strip trailing spaces from given name
      if ($n1 ne "") {
         $n1 .= " ";    # If there's a given name, add ONE trailing space to it
      }

      if ($lname eq "") {  # An empty surname is replaced with a constant
         $lname = $unknownsurname;
      }

      $n2 =~ s/^\s+//;  # Strip leading spaces from excess name
      $n2 =~ s/\s+$//;  # Strip trailing spaces from excess name
      if ($n2 ne "") {
         $n2 = " $n2";  # If there's an extra name, add ONE leading space to it
      }
   } else {
      $n1 = $nstr;      # The entire name is just a given name
      $lname = "";      # There is no surname
      $n2 = "";         # Nor is there any excess name
   }

   # Format the surname based on the $lfmt option
   $lname = wiki_text($lname, $lfmt);

   # Format the entire name based on the $nfmt option
   $nstr = wiki_text("$n1$lname$n2", $nfmt);

   # Return the modified string
   $nstr;
}

# Format a text string into Wiki format based on the passed-in format option
# Parm1 = String to convert
# Parm2 = Format option (bitmapped word)
sub wiki_text {
   # FMT_UPPERCASE?
   if (($_[1] & FMT_UPPERCASE) != 0) {
      $_[0] = uc $_[0];
   }
   # Bold?
   if (($_[1] & FMT_BOLD) != 0) {
      $_[0] = "**$_[0]**";
   }
   # Italics?
   if (($_[1] & FMT_ITALICS) != 0) {
      $_[0] = "//$_[0]//";
   }
   # Underline?
   if (($_[1] & FMT_UNDERLINE) != 0) {
      $_[0] = "__$_[0]__";
   }
   # Return the modified string
   $_[0];
}

# Format a text string into Wiki format based on the passed-in format option
# Parm1 = Name of link
# Parm1 = Text to use for link
sub format_link {
   "\[\[$fileprefix$_[1]\|$_[0]\]\]";
}

# Format the birth, death, and burial information for a specific person
# Parm1 = Xref id for Individual
sub format_data {
   my $itag = shift;
   my $outstr;

   # (If there is no data, then return an empty string)
   if ($birth{$itag} . $death{$itag} . $burial{$itag} ne "") {
      if ($private{$itag} != 0) {
         $outstr .= ", (Private)";
      } else {
         if ($birth{$itag} ne "") {
            $outstr = ", b. $birth{$itag}";
         }
         if ($death{$itag} ne "") {
            $outstr .= ", d. $death{$itag}";
         }
         if ($burial{$itag} ne "") {
            $outstr .= ", bur. $burial{$itag}";
         }
      }
   }
   # Return the formatted string to the caller
   $outstr;
}

# Output a NOTE record (if present)
# Parm1 = Xref id for the Note
# Parm2 = Linked to existing paragraph (0=No, 1=Yes)
sub output_note {
   my $nstr = shift;    # The note string (or xref id)
   my $linked = shift;  # Is this note part of an existing paragraph?

   if ($nstr eq "") {
      if ($linked != 0) {
         print FILE "\n";
      }
      return;
   }

   if ($nstr =~ /^\@(.*?)\@/) {
      if ($note{$1} ne "") {
         $nstr = $note{$1};
      }
   }

   # If one or more tabs follow a newline and are themselves followed by a number,
   #   make that line a list item.
   while ($nstr =~ /\n+(\t+)((\d+|[ivx]+))([\.\s])/) {
      my $n0 = "$`";
      my $tabs = $1;
      my $index = $2;
      my $dot = $4;
      my $n1 = "$'";
      $tabs =~ s/\t/  /g;
      $nstr = "$n0\n$tabs\* $index$dot$n1";
   }
   # All other tabs are turned into two spaces
   $nstr =~ s/\t/  /g;

   # Vertical tabs are turned into newlines
   $nstr =~ s/\x0b/\n/g;

   # Protect leading spaces on lines (except for list items)
   $nstr =~ s/\n( +)([^* ])/\n%%\1%%\2/g;

   # Three or more newlines is the same as two
   $nstr =~ s/\n\n\n+/\n\n/g;

   # Protect Wiki syntax characters
   $nstr =~ s/(\*{2,})/%%\1%%/g;
   $nstr =~ s/(\/{2,})/%%\1%%/g;
   $nstr =~ s/(\_{2,})/%%\1%%/g;

   # If there are http links enclosed in brackets (<>), turn them into wiki links
   # (Note: The "//" portion of the link would have been changed by the code above)
   while ($nstr =~ /<http:%%\/\/%%(.*?)>/) {
      $nstr = "$`\[\[http\:\/\/$1\]\]$'";
   }

   # If there are mailto links enclosed in brackets (<>), turn them into wiki links
   while ($nstr =~ /<mailto:(.*?)>/) {
      $nstr = "$`\[\[$1\]\]$'";
   }

   # Output the converted text
   print FILE "$nstr\n\n";
}
